#!/usr/bin/env python
# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.

from setuptools import setup


if __name__ == "__main__":
    setup()
