# Copyright Tephi contributors
#
# This file is part of Tephi and is released under the LGPL license.
# See COPYING and COPYING.LESSER in the root of the repository for full
# licensing details.
"""
Tephigram transform and isopleth constants.

"""

# TODO: Discover the meaning of the magic constant numbers.

CONST_CP = 1.01e3
CONST_K = 0.286
CONST_KELVIN = 273.15  # celsius to kelvin offset.
CONST_L = 2.5e6
CONST_MA = 300.0
CONST_RD = 287.0
CONST_RV = 461.0
